import datetime
import sys
import time
import traceback

from djangotools.common.date import CurrentDate


class CronTimer:

    def __init__(self):
        self.next_tick = None

    def __lt__(self, other):
        return self.next_tick < other.next_tick

    def next(self):
        raise NotImplementedError()

    def poll(self):
        if self.next_tick < time.time():
            self.next()
            return True
        return False

class CronTimerPeriod(CronTimer):
    NAME="period"
    def __init__(self, seconds):
        super().__init__()
        self.next_tick = time.time() + seconds
        self.next()

    def next(self):
        if self.next_tick<=time.time():
            self.next_tick = (self.next_tick or time.time()) + self.seconds


class CronTimerDaily(CronTimer):

    NAME="daily"
    def __init__(self, hour, minute=0, second=0):
        super().__init__()
        self.hour = hour
        self.minute = minute
        self.second = second
        self.next()

    def next(self):
        if not self.next_tick:
            next_tick = CurrentDate.now()
            next_tick = datetime.datetime(year=next_tick.year,
                                          month=next_tick.month,
                                          day=next_tick.day,
                                          hour=self.hour,
                                          minute=self.minute,
                                          second=self.second,
                                          tzinfo=next_tick.tzinfo)
            self.next_tick = next_tick.timestamp()
        if self.next_tick <= time.time():
            self.next_tick+=24*60*60


class Task:
    NAME =  None
    def __init__(self, *args, **kwargs):
        self.timers = []
        self.args = args
        self.kwargs = kwargs

    @property
    def next_tick(self):
        if self.timers:
            return min(x.next_tick for x in self.timers)
        return time.time()+3600*24

    def run(self, data):
        raise NotImplementedError()

    def poll(self, data):
        for timer in self.timers:
            if timer.poll():
                print(f"Execute task: {self.NAME}")
                try:
                    self.run(data)
                except Exception as err:
                    print(f"Erreur dans l'execution de la tache {self.NAME}", file=sys.stderr)
                    traceback.print_exc()

        return self.next_tick

    def add_timer(self, timer):
        self.timers.append(timer)

class CronManager:
    TASK_CLASS={

    }

    def __init__(self, resolution=3600):
        self.resolution = resolution
        self.tasks = []

    def add_task(self, task):
        if isinstance(task, Task):
            self.tasks.append(task)
        elif isinstance(task, (list, tuple, set)):
            for x in task: self.add_task(x)
        else:
            name = task.get("name")
            kwargs = task.get("kwargs") or {}
            args = task.get("args") or []
            timers = task.get("timers")
            task_classe = CronManager.TASK_CLASS.get(name)
            if task_classe is None:
                raise ValueError(f"Erreur la classe de tache '{name}' est inconnu")
            task = task_classe(*args, **kwargs)

            for type, *args in timers:
                timer_classe = None
                if type == "daily":
                    timer_classe = CronTimerDaily
                else:
                    timer_classe = CronTimerPeriod
                task.add_timer(timer_classe(*args))
            self.add_task(task)


    def schedule(self, data):
        while True:
            next = time.time() + 3600*24
            for task in self.tasks:
                next = min(next, task.poll(data))

            sleep = next - time.time()
            time.sleep(sleep)

    @staticmethod
    def register(task_calsse):
        CronManager.TASK_CLASS[task_calsse.NAME] = task_calsse


